%%
%% Install.oz
%%
%% install functors & other files from anywhere into the local or global cache

functor

import
   Error(exceptionToMessage messageToVirtualString /*printException*/)
   Open(file)
   OS(stat unlink)
   Pickle(loadWithHeader saveWithHeader)
   Property(get)
   Resolve%(make)
   System(printInfo showInfo showError)
   URL(toVirtualStringExtended)

   Toplevel
   XOS
   
export
   functors: InstallFunctors

define

   fun {ExceptionToVS Err}
      {Error.messageToVirtualString {Error.exceptionToMessage Err}}
   end
   
   `/`         = [XOS.path.sep]
   USER_HOME   = {Property.get 'user.home'}
   OZ_HOME     = {Property.get 'oz.home'}
   LOCALCACHE  = USER_HOME#`/`#".oz"#`/`#"cache"#`/`
   GLOBALCACHE = OZ_HOME#`/`#"cache"#`/`
   
   fun {MkDest Cache U}
      Cache#{URL.toVirtualStringExtended U how(cache:true)}
   end

   %%
   %% get file type (false if non-existing)
   %%
   fun {FileType File}
      try {OS.stat File}.type
      catch _ then false end
   end
   
   %%
   %% copy a file
   %%
   local
      SIZE = 10240
      %%
      fun {OpenRead File}
         {New Open.file init(name:File)}
      end
      fun {OpenWrite File}
         {New Open.file init(name:File flags:[write create truncate])}
      end
      %%
      proc {DoCopy F1 F2 Size}
         case {F1 read(list:$ size:Size)} of nil then skip
         [] S then
            {F2 write(vs:S)}
            {DoCopy F1 F2 Size}
         end
      end
   in
      proc {Copy File1 File2}
         F1 = {OpenRead File1} in
         try
            F2 = {OpenWrite File2} in
            try
               {DoCopy F1 F2 SIZE}
            finally
               {F2 close}
            end
         finally
            {F1 close}
         end
      end
   end

% components.cc: "save:open"             % 
% components.cc: "save:write"            % 
% components.cc: "save:malloc"           % 
% components.cc: "save:compressionlevel" %
      
% components.cc: "load:versionmismatch"  % muss abgefangen werden !!!
% components.cc: "load:nonpickle"        % muss abgefangen werden (?)
% components.cc: "load:read"             % ???

% components.cc: "bytesource:empty"      % muss abgefangen werden
% components.cc: "bytesource:crc"        % ERROR!

% components.cc: "getURL:thread"         % ???
% components.cc: "getURL:pipe"           % ???
% components.cc: "getURL:fork"           % ???
      
   %%
   %% file objects
   %%
   class File
      feat
         printInfo
         showInfo
         
      attr
         url:  unit
         file: unit
         new:  unit % @new==true implies that @file exists
         type: unit
         dest: unit
         
      meth init(U printInfo:PrintInfo showInfo:ShowInfo)
         url <- U
         self.printInfo = PrintInfo
         self.showInfo  = ShowInfo
      end
      %%
      meth download(Resolver<=Resolve)
         {self.showInfo "Localizing `"#@url#"'..."}
         case {Resolver.localize @url}
         of new(F) then new<-true  file<-F
            {self.showInfo "... saved at: "#F}
         [] old(F) then new<-false file<-F
            {self.showInfo "... found at: "#F}
         end
      end
      %%
      meth analyze
         try
            {self.showInfo "Analyzing `"#@file#"'..."}
            _#Val = {Pickle.loadWithHeader @file} in
            if {Functor.is Val}
            then type <- pickle('functor':true)
               {self.showInfo "... functor found"}
            else type <- pickle('functor':false)
               {self.showInfo "... pickle found"}
            end
         catch Err then
            case Err 
            of error(dp(generic 'bytesource:empty' _ _)...) then skip
               {self.showInfo "... ordinary file (non-pickle)."}
            [] error(dp(generic 'bytesource:crc' Msg _)...)   then skip
               {self.showInfo "*** WARNING: "#Msg#" in `"#@url#"'"}
               {self.showInfo " Treating it as ordinary file (non-pickle)..."}
            [] error(dp(generic 'load:versionmismatch' _ _)...) then
               raise abort(error(Err.1)) end
            [] system(url(load _ _) ...) then
               raise Err end
            else
               raise Err end
            end
         end
      end
      %%
      meth target(cache:Cache<=LOCALCACHE)
         dest <- {MkDest Cache @url}
         {self.printInfo "Checking `"#@dest#"'..."}
         %% if file exists and type\=reg --> ABORT INSTALLATION!
         Type   = {FileType @dest}
         Exists = (Type\=false)
      in
         if Exists
         then {self.showInfo " exists"}
         else {self.showInfo " non-existing"}
         end
         if Exists andthen (Type\=reg) then
            raise abort(nonReg @dest) end
         end
      end
      %%
      %% create missing directories
      %%
      meth mkDirs
         PathVS = @dest
         DirVS  = {XOS.path.dirname PathVS}
      in
         if {XOS.isDir DirVS}==false then
            {self.showInfo "Creating directory `"#DirVS#"'..."}
            {XOS.mkdir_p DirVS}
         end
      end
      %%
      meth install(compress:Compress<=0)
         %% if file exists and type==reg --> remove file first
         if {FileType @dest}==reg then
            {self.showInfo "Removing `"#@dest#"'..."}
            {OS.unlink @dest}
         end
         case @type of pickle('functor':IsFunctor) then
            Header#F = {Pickle.loadWithHeader @file} in
            if IsFunctor
            then {self.showInfo "Installing functor at `"#@dest#"'..."}
            else {self.showInfo "Installing pickle at `"#@dest#"'..."}
            end
            {Pickle.saveWithHeader F @dest Header Compress}
         else
            {self.showInfo "Installing file at `"#@dest#"'..."}
            {Copy @file @dest}
         end
      end
      %%
      meth cleanup
         if @new==true then
            {self.showInfo "Removing `"#@file#"'..."}
            {OS.unlink @file}
            file<-unit
            new <-unit
         end
      end
   end

   %%
   %% package objects
   %%
   local
      fun {NotUnit X} X\=unit end
   in
      fun {FilterUnit M}
         {Record.filter M NotUnit}
      end
   end
   
   class Package
      attr
         destCache:  unit
         components: nil
         
      meth init(URLs printInfo:PrintInfo showInfo:ShowInfo)
         components <- {Map URLs fun {$ U}
                                    {New File init(U printInfo:PrintInfo
                                                   showInfo:ShowInfo)}
                                 end}
      end
      %%
      meth download(Resolver<=unit) = M
         Download={FilterUnit M} in
         try
            {ForAll @components
             proc {$ C} {C Download} end}
         catch Err then
            Package, cleanup
            raise Err end
         end
      end
      %%
      meth analyze
         {ForAll @components
          proc {$ C} {C analyze} end}
      end
      %%
      meth target(cache:Cache<=unit) = M
         Target = {FilterUnit M} in
         {ForAll @components
          proc {$ C} {C Target} end}
      end
      %%
      meth mkDirs
         {ForAll @components
          proc {$ C} {C mkDirs} end}
      end
      %%
      meth install(compress:Compress<=unit) = M
         Install = {FilterUnit M} in
         {ForAll @components
          proc {$ C} {C Install} end}
      end
      %%
      meth cleanup
         {ForAll @components
          proc {$ C} {C cleanup} end}
      end
   end

   %%
   %% main
   %%
   
   Stages = stages("1. Select Cache"
                   "2. Download"
                   "3. Check Source"
                   "4. Check Target"
                   "5. Install Dirs"
                   "6. Install Files"
                   "7. Cleanup"
                   "8. Exit"
                  )
   
   Caches = caches('local':  LOCALCACHE
                   'global': GLOBALCACHE
                   default:  'local'
                  )
      
   proc {Install Title URLs ResolveVS DefaultCache Compress}
      
      MyCaches = {AdjoinAt Caches default DefaultCache}
      
      Top = {New Toplevel.'class' init(stages:Stages caches:MyCaches)}
      
      proc {Enter N}
         {Top enter(N)}
         {Top showInfo("Starting `"#Stages.N#"'.")}
      end
      proc {Leave N}
         {Top showInfo("Finished `"#Stages.N#"'.")}
         {Top leave(N)}
      end
      proc {DisplayError}
         {Top error}
      end
      
      DebugPrintInfo = proc {$ VS} {Top printInfo(VS)} end %System.printInfo
      DebugShowInfo  = proc {$ VS} {Top showInfo(VS)} end %System.showInfo
      ShowInfo       = proc {$ VS} {Top showInfo(VS)} end %System.showInfo
      ShowError      = proc {$ VS} {Top showError(VS)} end %System.showError
      PrintException = proc {$ Exc} {ShowError {ExceptionToVS Exc}} end

      {Top enableExit}
      {Top enter(1)}
      
      TitleVS = case Title of unit then {Length URLs}#" functors/files"
                [] VS then VS end
      
      {ShowInfo ''}
      {ShowInfo "Welcome to the Mozart Installer"}
      {ShowInfo "==============================="}
      {ShowInfo "\nThis programm will install"}
      {ShowInfo "\n               "#TitleVS}
      {ShowInfo "\nin your Oz cache. Check the selected destination,"}
      {ShowInfo "then click \"Install\" to start or \"Exit\" to abort."}
      {ShowInfo "\n\n\n\n\nThe following functors will be installed:"}
      {ForAll URLs ShowInfo}
      {ShowInfo "\nThe following resolver will be used for download:"}
      {ShowInfo ResolveVS}
      {ShowInfo "\nNOTE: In case of conflicts existing files will be replaced.\n"}
      DestinationCache = {Top cacheSelection($)}
      ExitSignal       = {Top exitSignal($)}
      {WaitOr DestinationCache ExitSignal}
      {Top leave(1)}
      {Top disableExit}
      if {IsDet ExitSignal} then
         raise exit end
      end
      
      %% init
      Pkg = {New Package init(URLs
                              printInfo: DebugPrintInfo
                              showInfo:  DebugShowInfo)}
   in
      try
         %% download
         {Enter 2}
         Resolver = {Resolve.make "download" vs(ResolveVS)}
      in
         {Pkg download(Resolver)}
         {Leave 2}
         %% analyze
         {Enter 3}
         {Pkg analyze}
         {Leave 3}
         %% look for existing files
         {Enter 4}
         {Pkg target(cache:DestinationCache)}
         {Leave 4}
         %% create all neccessary dirs
         {Enter 5}
         {Pkg mkDirs}
         {Leave 5}
         %% install files
         {Enter 6}
         {Pkg install(compress:Compress)}
         {Leave 6}
         {Top select(8)}
         {ShowInfo "\n*** INSTALLATION SUCCESSFUL ***\n"}
      catch Err then
         case Err
         of abort(nonReg Dest) then
            {ShowError "*** ERROR: `"#Dest#"' exists,"}
            {ShowError " but is not a regular file."}
         [] abort(E) then
            {PrintException E}
         else
            {PrintException Err}
         end
         {DisplayError}
         {Top select(8)}
         {ShowError "\n*** INSTALLATION ABORTED ***"}
         {ShowError "\nClick on the red button to see the error message."}
      finally
         %% cleanup
         {Enter 7}
         {Pkg cleanup}
         {Leave 7}
      end
      {Top enter(8)}
      {ShowInfo "\nClick on \"Exit\" to finish."}
      {Top enableExit}
      {Wait ExitSignal}
      raise exit end
   end

   %%
   %% public
   %%
   proc {InstallFunctors Args}
      Title     = {CondSelect Args title unit} 
      URLs      = {CondSelect Args urls nil}
      ResolveVS = {CondSelect Args resolve nil}
      Cache     = {CondSelect Args cache 'local'}
      Compress  = {CondSelect Args compress 0}
   in
      try {Install Title URLs ResolveVS Cache Compress}
      catch exit then skip end
   end
end
