%%
%% eXtensions of the OS system module
%%

functor
   
import
   Error(registerFormatter)
   Open(file)
   OS(stat system tmpnam unlink)
   Property(get)
   System(showInfo)
   Type(ask)
   
export
   path:  Path
   
   isDir: IsDir
   
   mkdir: MkDir
   rmdir: RmDir
   chgrp: ChGrp       
   chmod: ChMod
   
   getExec:    GetExec
   setExec:    SetExec
   setExecAll: SetExecAll
   
define

   VS2S = VirtualString.toString
   
   %%
   %% constants
   %%

   local
      OS      = {Property.get 'platform.os'} in
      WINDOWS = {List.isPrefix "win" {AtomToString OS}}
   end
   
   %%
   %% path analysis and synthesis 
   %%
   local
      Sep = if WINDOWS then &\\ else &/ end
      fun {SplitPath PathVS}
         {String.tokens {VS2S PathVS} Sep}
      end
      local
         fun {DoConcatPath PathL}
            case PathL of nil then nil
            [] P|Pr then Sep#P#{DoConcatPath Pr} end
         end
      in
         fun {ConcatPath P|Pr}
            P#{DoConcatPath Pr}
         end
      end
      fun {DirName PathVS}
         {Reverse {String.token {Reverse {VS2S PathVS}} Sep _ $}}
      end
      fun {BaseName PathVS}
         {Reverse {String.token {Reverse {VS2S PathVS}} Sep $ _}}
      end
   in
      Path = 'export'(sep:      Sep
                      split:    SplitPath
                      concat:   ConcatPath
                      dirname:  DirName
                      basename: BaseName
                     )
   end
   
   %%
   %% test -d (if a directory exists)
   %%
   fun {IsDir DirVS}
      try {OS.stat DirVS}.type==dir
      catch _ then false end
   end

   %%
   %% command line constructor functions for mkdir, rmdir, chgrp, chmod 
   %%
   Cmds = {Record.map cmds(mkdir:   if WINDOWS
                                    then "mkdir"
                                    else "\\mkdir -p"
                                    end
                           rmdir:   "\\rmdir"
                           chgrp:   "\\chgrp"
                           chmod:   "\\chmod")
           ByteString.make}
   local
      /*
      fun {EnumerateArgs Xs}
         case Xs of nil then ''
         [] X|Xr then ' '#X#{EnumerateArgs Xr} end
      end
      */
      CmdConstructors = 
      cmd(mkdir:   fun {$ DirVS}          Cmds.mkdir#' '#DirVS              end
          rmdir:   fun {$ DirVS}          Cmds.rmdir#' '#DirVS              end
          chgrp:   fun {$ GroupVS FileVS} Cmds.chgrp#' '#GroupVS#' '#FileVS end
          chmod:   fun {$ ModeVS  FileVS} Cmds.chmod#' '#ModeVS #' '#FileVS end
         )
      proc {FunApply F Args ?Result}
         {Procedure.apply F {Append Args [Result]}}
      end
   in
      fun {MkCmdLine Cmd Args}
         {FunApply CmdConstructors.Cmd Args} 
      end
   end
   
   %%
   %% allow overriding command execution (default: OS.system)
   %%
   local
      ExecProcs = {Record.map Cmds
                   fun {$ _} {NewCell OS.system} end}
   in
      fun {GetExec Cmd}
         {Access ExecProcs.Cmd}
      end
      proc {SetExec Cmd Proc2}
         {Type.ask.'procedure/2' Proc2}
         {Assign ExecProcs.Cmd Proc2}
      end
      proc {SetExecAll Proc2}
         {ForAll {Arity ExecProcs}
          proc {$ Cmd} {SetExec Cmd Proc2} end}
      end
   end
   
   %%
   %% execute commands & map errors to exceptions
   %%
   local
      proc {Do Cmd Args}
         E = {{GetExec Cmd} {MkCmdLine Cmd Args}} in
         if E>0 then
            {Exception.raiseError xos(Cmd E Args)}
         end
      end
   in
      proc {MkDir DirVS}
         {Do mkdir [DirVS]}
      end
      proc {RmDir DirVS} 
         {Do rmdir [DirVS]}
      end
      proc {ChGrp GroupVS FileVS}
         {Do chgrp [GroupVS FileVS]}
      end
      proc {ChMod ModeVS FileVS}
         {Do chmod [ModeVS FileVS]}
      end
   end

   %% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%% xos errors %%%%%%%%%%%%%%%%%%%%%%%%%
   {Error.registerFormatter xos
    fun {$ E}
       T = 'xos error'
       BugReport = 'please send a bug report to afranke@ags.uni-sb.de' 
    in
       case E of xos(Cmd Err Args) then
          case E 
          of xos(mkdir Err [DirVS]) then
             error(kind:T msg:'Cannot create directory'
                   items:[hint(l:'directory'  m:DirVS) unit
                          hint(l:'error code' m:Err)])
          [] xos(rmdir Err [DirVS]) then
             error(kind:T msg:'Cannot remove directory'
                   items:[hint(l:'directory'  m:DirVS) unit
                          hint(l:'error code' m:Err)])
          [] xos(chgrp Err [GroupVS FileVS]) then
             error(kind:T msg:'Cannot change group ownership'
                   items:[hint(l:'group'      m:GroupVS) 
                          hint(l:'file'       m:FileVS) unit
                          hint(l:'error code' m:Err)])
          [] xos(chmod Err [ModeVS FileVS]) then
             error(kind:T msg:'Cannot change permissions'
                   items:[hint(l:'mode'       m:ModeVS)
                          hint(l:'file'       m:FileVS) unit
                          hint(l:'error code' m:Err)])
          else
             error(kind:T msg:'XOS error'
                   items:[hint(l:'command'    m:Cmd)
                          hint(l:'args'       m:list(Args ' ')) unit
                          hint(l:'error code' m:Err) unit
                          line(BugReport)])
          end
       else
          error(kind:T msg:'Internal XOS error'
                items:[line(oz(E)) unit
                       line(BugReport)])
       end
    end}
   %% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
end

